      PROGRAM SPECNSE
      IMPLICIT NONE
      
      INTEGER sx,sy,syp1,wsize
      PARAMETER(sx=64,sy=64,syp1=sy+1,wsize=sx+20)
      REAL*8 pi
      PARAMETER(pi=3.14159265358979323846264338327950288D+0)
      
C general variables

      INTEGER  n,m,mm1,mp1,i,j,k,kk
      REAL*8   xlen,t1,t2,t3,t4
      REAL*8   x(0:sx),y(0:sy),nu,re,cadjust

C Chebyshev collocation variables and factorizations

      REAL*8  D1(0:sy,0:sy),D2(0:sy,0:sy)
      REAL*8  Q(sy,sy),Qinv(sy,sy),R(sy,sy),Rinv(sy,sy)
      REAL*8  c0_m,c0_p,cM_m,cM_p,e,b0(sy),bM(sy)

C variables for VFFTPK

      REAL*8 fm1(sx),fm2(sx),wsave(wsize),fftwork(0:sy,0:sx)

C evals matrices for Poisson equations

      REAL*8 evmatD(sy,sx),evmatN(sy,sx)

C physical variables (true solution and force terms)  

      REAL*8 tu(0:sy,0:sx),tv(0:sy,0:sx),tp(0:sy,0:sx)
      REAL*8 ftu(0:sy,0:sx),ftv(0:sy,0:sx),ftp(0:sy,0:sx)
      REAL*8 fu1(0:sy,0:sx),fu2(0:sy,0:sx),fu3(0:sy,0:sx)
      REAL*8 fv1(0:sy,0:sx),fv2(0:sy,0:sx),fv3(0:sy,0:sx)
      REAL*8 fu1_x(0:sy,0:sx),fu2_x(0:sy,0:sx),fu3_x(0:sy,0:sx)
      REAL*8 fv1_y(0:sy,0:sx),fv2_y(0:sy,0:sx),fv3_y(0:sy,0:sx)      

C time variables

      INTEGER nts
      REAL*8 final_time,tfinal,cfl,dt,tf1,tf2,tf3
      REAL*8 ft1,ft2,ft3,ctime,ftime
      
C time stepping variables (Fourier transformed)

      REAL*8 u1(0:sy,0:sx),v1(0:sy,0:sx),p1(0:sy,0:sx)
      REAL*8 u2(0:sy,0:sx),v2(0:sy,0:sx),p2(0:sy,0:sx)
      REAL*8 unew(0:sy,0:sx),vnew(0:sy,0:sx),pnew(0:sy,0:sx)   
      REAL*8 NLT_u1(0:sy,0:sx),NLT_u2(0:sy,0:sx)  
      REAL*8 NLT_v1(0:sy,0:sx),NLT_v2(0:sy,0:sx)
      REAL*8 NLT_unew(0:sy,0:sx),NLT_vnew(0:sy,0:sx)
      REAL*8 py_m(0:sx),py_p(0:sx)   

C aux matrix variables

      REAL*8 m1(0:sy,0:sx),m2(0:sy,0:sx),m3(0:sy,0:sx),m4(0:sy,0:sx)


      CALL SETUP(n,m,x,xlen,y,re,nu,final_time,tfinal,cfl,dt,nts,
     &           D1,D2,Q,Qinv,R,Rinv,evmatD,evmatN,fm1,fm2,wsave,
     &           c0_m,c0_p,cM_m,cM_p,e,b0,bM,sx,sy,wsize)
      mm1=m-1
      mp1=m+1

C compute exact solution and forcing functions. ftu, ftv and ftp
C are Fourier transformed versions of fu, fv, and fp. ALL of the 
C force functions are returned transformed.

      CALL ACCFUNCS(n,m,x,y,nu,tu,tv,tp,fu1,fu2,fu3,ftu,ftv,ftp,
     &              fv1,fv2,fv3,fu1_x,fu2_x,fu3_x,fv1_y,fv2_y,fv3_y,
     &              fftwork,wsave,sx,sy)

C set initial data at t0 and t0-dt, including nonlinear terms
C TRUE SOLUTION: u=tf1(tfinal)*tu, v=tf1(tfinal)*tv, p=tf1(tfinal)*tp 

      ctime=0.0D+0
      t1=tf1(-dt)
      t2=tf1(ctime)
      DO j=1,n
         DO i=0,m
            u1(i,j)=t1*ftu(i,j)
            u2(i,j)=t2*ftu(i,j)
            v1(i,j)=t1*ftv(i,j)
            v2(i,j)=t2*ftv(i,j) 
            p1(i,j)=t1*ftp(i,j)
            p2(i,j)=t2*ftp(i,j)            
            NLT_u1(i,j)=t1*fu3(i,j)
            NLT_u2(i,j)=t2*fu3(i,j)
            NLT_v1(i,j)=t1*fv3(i,j)
            NLT_v2(i,j)=t2*fv3(i,j)
         END DO
      END DO


CCCCCCCCCCCCCCCC
C MAIN TIME LOOP
CCCCCCCCCCCCCCCC

      DO kk=1,nts

         ctime=DBLE(kk-1)*dt
         IF (MOD(kk,500) .EQ. 0) THEN
            WRITE(6,*) 'step :',kk,' time ;',ctime
         END IF

C update velocities

         ftime=ctime+dt/2.0D+0
         ft1=tf1(ftime)
         ft2=tf2(ftime)
         ft3=tf3(ftime)

C compute m1=lap(u2), m2=lap(v2), m3=Dx(3*p2-p1)/2, m4=Dy(3*p2-p1)/2

         CALL COMP_VTERMS(m,n,m1,m2,m3,m4,u1,u2,v1,v2,p1,p2,
     &                    D1,D2,fm1,fm2,sx,sy)

C compute RHS of velocity equations, and solve

         DO j=1,n
            DO i=1,mm1
               unew(i,j)=u2(i,j)+dt*(0.5D+0*nu*m1(i,j)-m3(i,j)
     &                   -(3.0D+0*NLT_u2(i,j)-NLT_u1(i,j))/2.0D+0
     &                   +ft1*fu1(i,j)+ft2*fu2(i,j)+ft3*fu3(i,j)) 
               vnew(i,j)=v2(i,j)+dt*(0.5D+0*nu*m2(i,j)-m4(i,j)
     &                   -(3.0D+0*NLT_v2(i,j)-NLT_v1(i,j))/2.0D+0
     &                   +ft1*fv1(i,j)+ft2*fv2(i,j)+ft3*fv3(i,j))
            END DO
         END DO              

         DO j=1,n
            DO i=1,mm1
               t1=0.0D+0
               t2=0.0D+0
               DO k=1,mm1
                  t1=t1+Qinv(i,k)*unew(k,j)
                  t2=t2+Qinv(i,k)*vnew(k,j)               
               END DO
               m1(i,j)=t1/(1.0D+0-nu*dt*evmatD(i,j)/2.0D+0)
               m2(i,j)=t2/(1.0D+0-nu*dt*evmatD(i,j)/2.0D+0)            
            END DO
         END DO

         DO j=1,n-1
            DO i=1,mm1
               t1=0.0D+0
               t2=0.0D+0
               DO k=1,mm1
                  t1=t1+Q(i,k)*m1(k,j)
                  t2=t2+Q(i,k)*m2(k,j)               
               END DO
               unew(i,j)=t1
               vnew(i,j)=t2
            END DO
         END DO   

C set homogeneous BC (all Fourier coeffs zero!), and compute new NLT

         DO i=1,mm1
            unew(i,n)=0.0D+0
            vnew(i,n)=0.0D+0
         END DO      
         DO j=1,n
            unew(0,j)=0.0D+0
            unew(m,j)=0.0D+0
            vnew(0,j)=0.0D+0
            vnew(m,j)=0.0D+0
         END DO

         CALL COMP_NLT(m,n,unew,vnew,NLT_unew,NLT_vnew,
     &                 m1,m2,m3,m4,D1,fm1,fftwork,wsave,sx,sy)

C compute RHS of PPE, accounting for force terms

         ftime=ctime+dt
         ft1=tf1(ftime)
         ft2=tf2(ftime)
         ft3=tf3(ftime)  
       
C j=1 and j=n
 
         DO i=1,mm1
            t1=0.0D+0
            t2=0.0D+0
            DO k=0,m 
               t1=t1+D1(i,k)*NLT_vnew(k,1)    
               t2=t2+D1(i,k)*NLT_vnew(k,n) 
            END DO
            pnew(i,1)=-t1
     &                +ft1*fu1_x(i,1)+ft2*fu2_x(i,1)+ft3*fu3_x(i,1) 
     &                +ft1*fv1_y(i,1)+ft2*fv2_y(i,1)+ft3*fv3_y(i,1)
            pnew(i,n)=-t2
     &                +ft1*fu1_x(i,n)+ft2*fu2_x(i,n)+ft3*fu3_x(i,n) 
     &                +ft1*fv1_y(i,n)+ft2*fv2_y(i,n)+ft3*fv3_y(i,n)
         END DO

C j=2,...,n-1
         
         DO j=2,n-2,2
            DO i=1,mm1
               t1=0.0D+0
               t2=0.0D+0
               DO k=0,m 
                  t1=t1+D1(i,k)*NLT_vnew(k,j)    
                  t2=t2+D1(i,k)*NLT_vnew(k,j+1) 
               END DO
               pnew(i,j)=-(fm1(j+1)*NLT_unew(i,j+1)+t1)
     &               +ft1*fu1_x(i,j)+ft2*fu2_x(i,j)+ft3*fu3_x(i,j) 
     &               +ft1*fv1_y(i,j)+ft2*fv2_y(i,j)+ft3*fv3_y(i,j)
               pnew(i,j+1)=-(fm1(j)*NLT_unew(i,j)+t2)
     &               +ft1*fu1_x(i,j+1)+ft2*fu2_x(i,j+1)+ft3*fu3_x(i,j+1) 
     &               +ft1*fv1_y(i,j+1)+ft2*fv2_y(i,j+1)+ft3*fv3_y(i,j+1)
            END DO
         END DO

C compute Neumann BC: dp/dy =-nu*(-v_xx+u_yx)=-nu*u_yx since v_xx=0

         DO j=2,n-2,2
            t1=0.0D+0
            t2=0.0D+0
            t3=0.0D+0
            t4=0.0D+0
            DO k=0,m
               t1=t1+D1(0,k)*unew(k,j)
               t2=t2+D1(m,k)*unew(k,j)
               t3=t3+D1(0,k)*unew(k,j+1)
               t4=t4+D1(m,k)*unew(k,j+1)
            END DO
            py_p(j)=-nu*fm1(j+1)*t3
     &              +ft1*fv1(0,j)+ft2*fv2(0,j)+ft3*fv3(0,j)                   
            py_p(j+1)=-nu*fm1(j)*t1
     &              +ft1*fv1(0,j+1)+ft2*fv2(0,j+1)+ft3*fv3(0,j+1)
            py_m(j)=-nu*fm1(j+1)*t4
     &              +ft1*fv1(m,j)+ft2*fv2(m,j)+ft3*fv3(m,j)                   
            py_m(j+1)=-nu*fm1(j)*t2
     &              +ft1*fv1(m,j+1)+ft2*fv2(m,j+1)+ft3*fv3(m,j+1)
         END DO
         py_p(0)=ft1*fv1(0,0)+ft2*fv2(0,0)+ft3*fv3(0,0)
         py_p(n)=ft1*fv1(0,n)+ft2*fv2(0,n)+ft3*fv3(0,n)
         py_m(0)=ft1*fv1(m,0)+ft2*fv2(m,0)+ft3*fv3(m,0)
         py_m(n)=ft1*fv1(m,n)+ft2*fv2(m,n)+ft3*fv3(m,n)

C adjust RHS of PPE to account for BC

         DO j=1,n
            DO i=1,mm1
               pnew(i,j)=pnew(i,j)-((c0_m*py_m(j)+c0_p*py_p(j))*D2(i,0)
     &                          +(cM_m*py_m(j)+cM_p*py_p(j))*D2(i,m))/e
            END DO
         END DO

C solve PPE

         DO j=1,n
            DO i=1,mm1
               t1=0.0D+0
               DO k=1,mm1
                  t1=t1+Rinv(i,k)*pnew(k,j)
               END DO
               m1(i,j)=t1/(evmatN(i,j))
            END DO
         END DO
         m1(1,1)=0.0D+0

         DO j=1,n
            DO i=1,mm1
               t1=0.0D+0
               DO k=1,mm1
                  t1=t1+R(i,k)*m1(k,j)
               END DO
               pnew(i,j)=t1
            END DO
         END DO   

C recover pressure boundary values and set n coeff to 0

         DO j=1,n-1
            t1=0.0D+0
            t2=0.0D+0
            DO i=1,mm1
               t1=t1+b0(i)*pnew(i,j)
               t2=t2+bM(i)*pnew(i,j)
            END DO
            pnew(0,j)=(t1+c0_m*py_m(j)+c0_p*py_p(j))/e
            pnew(m,j)=(t2+cM_m*py_m(j)+cM_p*py_p(j))/e
         END DO    

         DO i=0,m
            pnew(i,n)=0.0D+0
         END DO 

C update all variables

         DO j=1,n
            DO i=0,m
               u1(i,j)=u2(i,j)
               v1(i,j)=v2(i,j)
               p1(i,j)=p2(i,j) 
               NLT_u1(i,j)=NLT_u2(i,j)     
               NLT_v1(i,j)=NLT_v2(i,j)
            END DO
         END DO
         DO j=1,n
            DO i=0,m
               u2(i,j)=unew(i,j)
               v2(i,j)=vnew(i,j)
               p2(i,j)=pnew(i,j) 
               NLT_u2(i,j)=NLT_unew(i,j)     
               NLT_v2(i,j)=NLT_vnew(i,j)
            END DO
         END DO

CCCCCCCCCCCCCCC
C END TIME LOOP
CCCCCCCCCCCCCCC

      END DO 


C recover physical values, set periodicity, and check error
 
      CALL VRFFTB(mm1,n,unew(1,1),fftwork,syp1,wsave)      
      CALL VRFFTB(mm1,n,vnew(1,1),fftwork,syp1,wsave)   
      CALL VRFFTB(m+1,n,pnew(0,1),fftwork,syp1,wsave) 
      
      DO j=1,n
         unew(0,j)=0.0D+0
         unew(m,j)=0.0D+0
         vnew(0,j)=0.0D+0
         vnew(m,j)=0.0D+0
      END DO
      DO i=0,m
         unew(i,0)=unew(i,n)
         vnew(i,0)=vnew(i,n)
         pnew(i,0)=pnew(i,n)
      END DO
 
      ctime=tf1(tfinal)
      WRITE(6,*) ' '
      WRITE(6,*) 'Final time : ',tfinal
      WRITE(6,*) ' '      

      t1=0.0D+0
      DO j=0,n
         DO i=0,m
            t2=DABS(unew(i,j)-ctime*tu(i,j))
            IF (t2 .GT. t1) THEN
               t1=t2
            END IF
         END DO
      END DO
      WRITE(6,*) 'L-inf error u : ',t1 

      t1=0.0D+0
      DO j=0,n
         DO i=0,m
            t2=DABS(vnew(i,j)-ctime*tv(i,j))
            IF (t2 .GT. t1) THEN
               t1=t2
            END IF
         END DO
      END DO
      WRITE(6,*) 'L-inf error v : ',t1       

      cadjust=ctime*tp(m/2,n/2)-pnew(m/2,n/2)
      t1=0.0D+0
      DO j=0,n
         DO i=0,m
            t2=DABS((pnew(i,j)+cadjust)-ctime*tp(i,j))
            IF (t2 .GT. t1) THEN
               t1=t2
            END IF
         END DO
      END DO
      
      WRITE(6,*) 'L-inf error p : ',t1        
      WRITE(6,*) ' '

      STOP
      END
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      SUBROUTINE SETUP(n,m,x,xlen,y,re,nu,final_time,tfinal,cfl,dt,nts,
     &                 D1,D2,Q,Qinv,R,Rinv,evmatD,evmatN,fm1,fm2,wsave,
     &                 c0_m,c0_p,cM_m,cM_p,e,b0,bM,sx,sy,wsize)
      IMPLICIT NONE

      INTEGER lsize,lwork
      PARAMETER(lsize=64,lwork=25*lsize)
      REAL*8 pi
      PARAMETER(pi=3.14159265358979323846264338327950288D+0)

      INTEGER n,m,sx,sy,wsize,nts
      REAL*8  re,nu
      REAL*8  xlen,x(0:sx),y(0:sy),c0_m,c0_p,cM_m,cM_p,e,b0(sy),bM(sy)
      REAL*8  D1(0:sy,0:sy),D2(0:sy,0:sy)
      REAL*8  Q(sy,sy),Qinv(sy,sy),R(sy,sy),Rinv(sy,sy)
      REAL*8  evmatD(sy,sx),evmatN(sy,sx)
      REAL*8  fm1(sx),fm2(sx),wsave(wsize)
      REAL*8  final_time,tfinal,cfl,dt

C local variables
    
      INTEGER i,j,mm1,info,ipiv(lsize),minindex
      REAL*8  dx,dy,vl,t1
      REAL*8  tmat(lsize,lsize),work(lwork)
      REAL*8  wr(lsize),wi(lsize)

      WRITE(6,*) ' '
      WRITE(6,*) 'Fourier-Chebshev solver on [0,2*pi]x[-1,1] '
      WRITE(6,*) ' '
      WRITE(6,*) 'Enter n (x) and m (y) :'
      READ(5,*) n,m
      mm1=m-1
      WRITE(6,*) 'Enter Reynolds number :'
      READ(5,*) re
      nu=1.0D+0/re
      WRITE(6,*) 'Enter final time and cfl number :'
      READ(5,*) final_time,cfl

C      WRITE(6,*) 'Enter x-length:'
C      READ(5,*)  xlen
C to test solver take xlen=2*pi
      xlen=2.0D+0*pi

      IF ((sy .LT. m) .OR. (sx .LT. n)) THEN
         WRITE(6,*) 'IN SETUP: n or m  are too large!!'
         STOP
      END IF     

       IF (lsize .LT. m-1) THEN
         WRITE(6,*) 'IN SETUP: lsize is too small!!'
         WRITE(6,*) 'Must be at least ',m-1
         STOP
      END IF    

C compute Chebyshev differentiation matrices

      CALL CHEB(y,D1,D2,m,sy)

C variables for Neumann problem

      c0_m=-D1(0,m)
      c0_p=D1(m,m)
      cM_p=-D1(m,0)
      cM_m=D1(0,0)
      e=c0_p*cM_m-c0_m*cM_p
      DO j=1,mm1
         b0(j)=-c0_p*D1(0,j)-c0_m*D1(m,j) 
         bM(j)=-cM_m*D1(m,j)-cM_p*D1(0,j)
      END DO

C Fourier multiples and initialization

      dx=xlen/DBLE(n)
      DO j=0,n
         x(j)=DBLE(j)*dx
      END DO
      fm1(1)=0.0D+0
      fm2(1)=0.0D+0
      DO J=2,n-2,2
         fm1(j  )=pi*DBLE(j)/xlen
         fm1(j+1)=-fm1(j)
         fm2(j  )=-(pi*DBLE(j)/xlen)**2
         fm2(j+1)=fm2(j)
      END DO
      fm1(n)=0.0D+0
C fm2(n) is reset to 0 below!
      fm2(n)=1.0D+0

      CALL VRFFTI(n,wsave)

C Dirichlet problem decomposition and multiples matrix

      DO j=1,mm1
         DO i=1,mm1
            tmat(i,j)=D2(i,j)
         END DO
      END DO      

      CALL DGEEV('N','V',mm1,tmat,lsize,wr,wi,vl,1,Q,sy,
     &           work,lwork,info)

      DO j=1,mm1
         DO i=1,mm1
            Qinv(i,j)=Q(i,j)
         END DO
      END DO

      CALL DGETRF(mm1,mm1,Qinv,sy,ipiv,info)
      CALL DGETRI(mm1,Qinv,sy,ipiv,work,lwork,info)

      DO j=1,n
         DO i=1,mm1
            evmatD(i,j)=wr(i)+fm2(j)
         END DO
      END DO

      
C Neumann problem decomposition and multiples matrix

      DO j=1,mm1
         DO i=1,mm1
            tmat(i,j)=D2(i,j)+(D2(i,0)*b0(j)+D2(i,m)*bM(j))/e
         END DO
      END DO

      CALL DGEEV('N','V',mm1,tmat,lsize,wr,wi,vl,1,R,sy,
     &            work,lwork,info)

C find minimum (should be 0) eval position and switch evals and evectors

      t1=wr(1)
      minindex=1
      DO j=2,mm1
         IF (DABS(wr(j)) .LT. DABS(t1)) THEN
            minindex=j
            t1=wr(j)
         END IF
      END DO

      wr(minindex)=wr(1)
      wr(1)=0.0D+0
      DO j=1,mm1
         t1=R(j,1)
         R(j,1)=R(j,minindex)
         R(j,minindex)=t1
      END DO

      DO j=1,mm1
         DO i=1,mm1
            Rinv(i,j)=R(i,j)
         END DO
      END DO

      CALL DGETRF(mm1,mm1,Rinv,sy,ipiv,info)
      CALL DGETRI(mm1,Rinv,sy,ipiv,work,lwork,info)

      DO j=1,n
         DO i=1,mm1
            evmatN(i,j)=wr(i)+fm2(j)
         END DO
      END DO
      evmatN(1,1)=1.0D+0
      fm2(n)=0.0D+0

C time stepping 

      dy=y(0)-y(1)
      dt=cfl*DMIN1(dx,dy)
      nts=IDINT(final_time/dt)
      tfinal=DBLE(nts)*dt

      WRITE(6,*) ' '
      WRITE(6,*) 'tfinal, nts, dt :',tfinal,nts,dt
      WRITE(6,*) ' '

      RETURN
      END
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      SUBROUTINE ACCFUNCS(n,m,x,y,nu,tu,tv,tp,fu1,fu2,fu3,ftu,ftv,ftp,
     &                    fv1,fv2,fv3,fu1_x,fu2_x,fu3_x,fv1_y,fv2_y,
     &                    fv3_y,fftwork,wsave,sx,sy)
      IMPLICIT NONE

      INTEGER n,m,sx,sy
      REAL*8 x(0:sx),y(0:sy),nu
      REAL*8 tu(0:sy,0:sx),tv(0:sy,0:sx),tp(0:sy,0:sx)
      REAL*8 ftu(0:sy,0:sx),ftv(0:sy,0:sx),ftp(0:sy,0:sx)
      REAL*8 fu1(0:sy,0:sx),fu2(0:sy,0:sx),fu3(0:sy,0:sx)
      REAL*8 fv1(0:sy,0:sx),fv2(0:sy,0:sx),fv3(0:sy,0:sx)
      REAL*8 fu1_x(0:sy,0:sx),fu2_x(0:sy,0:sx),fu3_x(0:sy,0:sx)
      REAL*8 fv1_y(0:sy,0:sx),fv2_y(0:sy,0:sx),fv3_y(0:sy,0:sx)
      REAL*8 fftwork(0:sy,0:sx),wsave(sx)  

C local variables

      INTEGER i,j
      REAL*8 pi
      PARAMETER(pi=3.14159265358979323846264338327950288D+0)

C tu = 2y*(1-y^2)*sin(x)
C tv = (1-y^2)^2*cos(x)
C tp = exp(y)*sin(x)

      DO j=0,n
         DO i=0,m
            tu(i,j)=4.0D+0*y(i)*(1-y(i)**2)*DSIN(x(j))
            tv(i,j)=(1-y(i)**2)**2*DCOS(x(j))
            tp(i,j)=DEXP(y(i))*DSIN(x(j))
         END DO
      END DO

C fu1 = tpx - nu*(tuxx+tuyy);
C fu2 = tu;
C fu3 = tu.*tux + tv.*tuy;

      DO j=0,n
         DO i=0,m
            fu1(i,j)=DEXP(y(i))*DCOS(x(j))+28.0D+0*nu*y(i)*DSIN(x(j))
     &               -4.0D+0*nu*DSIN(x(j))*y(i)**3
            fu2(i,j)=tu(i,j)
            fu3(i,j)=4.0D+0*DSIN(x(j))*DCOS(x(j))*
     &               (-y(i)**2-y(i)**4+y(i)**6+1.0D+0)
            fu1_x(i,j)=-DEXP(y(i))*DSIN(x(j))+28.0D+0*nu*y(i)*DCOS(x(j))
     &                 -4.0D+0*nu*DCOS(x(j))*y(i)**3
            fu2_x(i,j)=4.0D+0*y(i)*(1-y(i)**2)*DCOS(x(j))
            fu3_x(i,j)=4.0D+0*(2.0D+0*DCOS(x(j))**2-1.0D+0)*
     &                 (-y(i)**2-y(i)**4+y(i)**6+1.0D+0)
         END DO
      END DO    

C fv1 = tpy - nu*(tvxv+tvyy);
C fv2 = tv;
C fv3 = tu.*tvx + tv.*tvy;      

      DO j=0,n
         DO i=0,m
            fv1(i,j)=DEXP(y(i))*DSIN(x(j))+5.0D+0*nu*DCOS(x(j))
     &              -14.0D+0*nu*DCOS(x(j))*y(i)**2+nu*DCOS(x(j))*y(i)**4
            fv2(i,j)=tv(i,j)
            fv3(i,j)=-4.0D+0*y(i)+12.0D+0*y(i)**3-12.0D+0*y(i)**5+
     &                4.0D+0*y(i)**7
            fv1_y(i,j)=DEXP(y(i))*DSIN(x(j))-28.0D+0*nu*y(i)*DCOS(x(j))
     &                 +4.0D+0*nu*DCOS(x(j))*y(i)**3
            fv2_y(i,j)=4.0D+0*y(i)*(-1.0D+0+y(i)**2)*DCOS(x(j))
            fv3_y(i,j)=-4.0D+0+36.0D+0*y(i)**2-60.0D+0*y(i)**4
     &                 +28.0D+0*y(i)**6
         END DO
      END DO

C copy tu, tv and tp to ftu, ftv and ftp (Fourier versions)

      DO j=0,n
         DO i=0,m
            ftu(i,j)=tu(i,j)
            ftv(i,j)=tv(i,j)
            ftp(i,j)=tp(i,j)
         END DO
      END DO

C transform physical variables in periodic direction (x)

      CALL VRFFTF(m+1,n,ftu(0,1),fftwork,sy+1,wsave)
      CALL VRFFTF(m+1,n,ftv(0,1),fftwork,sy+1,wsave)
      CALL VRFFTF(m+1,n,ftp(0,1),fftwork,sy+1,wsave)            

C transform force terms in periodic direction (x)

      CALL VRFFTF(m+1,n,fu1(0,1),fftwork,sy+1,wsave)
      CALL VRFFTF(m+1,n,fu2(0,1),fftwork,sy+1,wsave)          
      CALL VRFFTF(m+1,n,fu3(0,1),fftwork,sy+1,wsave)
      CALL VRFFTF(m+1,n,fu1_x(0,1),fftwork,sy+1,wsave)
      CALL VRFFTF(m+1,n,fu2_x(0,1),fftwork,sy+1,wsave)          
      CALL VRFFTF(m+1,n,fu3_x(0,1),fftwork,sy+1,wsave)
      CALL VRFFTF(m+1,n,fv1(0,1),fftwork,sy+1,wsave)
      CALL VRFFTF(m+1,n,fv2(0,1),fftwork,sy+1,wsave)          
      CALL VRFFTF(m+1,n,fv3(0,1),fftwork,sy+1,wsave)
      CALL VRFFTF(m+1,n,fv1_y(0,1),fftwork,sy+1,wsave)
      CALL VRFFTF(m+1,n,fv2_y(0,1),fftwork,sy+1,wsave)          
      CALL VRFFTF(m+1,n,fv3_y(0,1),fftwork,sy+1,wsave)

      RETURN
      END

!!!!!!!!!!!!!!!!!!!!!
      SUBROUTINE COMP_VTERMS(m,n,m1,m2,m3,m4,u1,u2,v1,v2,p1,p2,
     &                       D1,D2,fm1,fm2,sx,sy)
      IMPLICIT NONE
      
      INTEGER m,n,sx,sy
      REAL*8  m1(0:sy,0:sx),m2(0:sy,0:sx),m3(0:sy,0:sx),m4(0:sy,0:sx)
      REAL*8  u1(0:sy,0:sx),u2(0:sy,0:sx),v1(0:sy,0:sx),v2(0:sy,0:sx)      
      REAL*8  p1(0:sy,0:sx),p2(0:sy,0:sx)
      REAL*8  D1(0:sy,0:sy),D2(0:sy,0:sy),fm1(sx),fm2(sx)
      
C local variables

      INTEGER i,j,k,mm1
      REAL*8  t1,t2,t3,t4
            
      mm1=m-1
C
C m1=lap(u2),m2=lap(v2),m3=Dx(3*p2-p1)/2 
C
C j=1 & j=n

      DO i=1,mm1
         t1=0.0D+0
         t2=0.0D+0
         t3=0.0D+0
         t4=0.0D+0
         DO k=1,mm1
            t1=t1+D2(i,k)*u2(k,1)
            t2=t2+D2(i,k)*u2(k,n)
            t3=t3+D2(i,k)*v2(k,1)
            t4=t4+D2(i,k)*v2(k,n)
         END DO
         m1(i,1)=t1
         m1(i,n)=t2
         m2(i,1)=t3
         m2(i,n)=t4
         m3(i,1)=0.0D+0
         m3(i,n)=0.0D+0
      END DO

C j=2,...,n-1

      DO j=2,n-2,2
         DO i=1,mm1
            t1=0.0D+0
            t2=0.0D+0
            t3=0.0D+0
            t4=0.0D+0
            DO k=1,mm1
               t1=t1+D2(i,k)*u2(k,j)
               t2=t2+D2(i,k)*u2(k,j+1)
               t3=t3+D2(i,k)*v2(k,j)
               t4=t4+D2(i,k)*v2(k,j+1)
            END DO
            m1(i,j)=t1+fm2(j)*u2(i,j)
            m1(i,j+1)=t2+fm2(j+1)*u2(i,j+1)
            m2(i,j)=t3+fm2(j)*v2(i,j)
            m2(i,j+1)=t4+fm2(j+1)*v2(i,j+1)
            m3(i,j)=fm1(j+1)*(3.0D+0*p2(i,j+1)-p1(i,j+1))/2.0D+0
            m3(i,j+1)=fm1(j)*(3.0D+0*p2(i,j)-p1(i,j))/2.0D+0                        
         END DO
      END DO

C m4=Dy(3*p2-p1)/2; include k=0 and k=m terms in D1 mult!!

      DO j=1,n
         DO i=1,mm1
            t1=0.0D+0
            DO k=0,m
               t1=t1+D1(i,k)*(3.0D+0*p2(k,j)-p1(k,j))/2.0D+0
            END DO
            m4(i,j)=t1
         END DO
      END DO                  

      RETURN
      END   
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      SUBROUTINE COMP_NLT(m,n,u,v,nlu,nlv,m1,m2,m3,m4,
     &                    D1,fm1,fftwork,wsave,sx,sy) 
      IMPLICIT NONE
      
      INTEGER m,n,sx,sy
      REAL*8  u(0:sy,0:sx),v(0:sy,0:sx),nlu(0:sy,0:sx),nlv(0:sy,0:sx)  
      REAL*8  m1(0:sy,0:sx),m2(0:sy,0:sx),m3(0:sy,0:sx),m4(0:sy,0:sx)  
      REAL*8  D1(0:sy,0:sy),fm1(sx),fftwork(0:sy,0:sx),wsave(sx)

C local variables

      INTEGER i,j,k,mm1,syp1
      REAL*8  t1,t2
      
      mm1=m-1
      syp1=sy+1
 
C Note: for future purposes DO NOT assume homogeneous BC.  Thus, when
C taking the derivative in y (Chebyshev) one must include i=0 and i=m.
      
C m1=physcial(u), m2=physical(v) : copy and tansform 

      DO j=1,n
         DO i=0,m
            m1(i,j)=u(i,j)
            m2(i,j)=v(i,j)
         END DO
      END DO
      CALL VRFFTB(m+1,n,m1(0,1),fftwork,syp1,wsave)
      CALL VRFFTB(m+1,n,m2(0,1),fftwork,syp1,wsave) 

C m3=physical(u_x), m4=physical(u_y)

C j=1 and j=n

      DO i=0,m
         t1=0.0D+0
         t2=0.0D+0
         DO k=0,m
            t1=t1+D1(i,k)*u(k,1)
            t2=t2+D1(i,k)*u(k,n) 
         END DO
         m3(i,1)=0.0D+0
         m3(i,n)=0.0D+0
         m4(i,1)=t1
         m4(i,n)=t2
      END DO
      
C j=2,...,n-1
      
      DO j=2,n-2,2
         DO i=0,m
            t1=0.0D+0
            t2=0.0D+0
            DO k=0,m
               t1=t1+D1(i,k)*u(k,j)
               t2=t2+D1(i,k)*u(k,j+1) 
            END DO
            m3(i,j)=fm1(j+1)*u(i,j+1)
            m3(i,j+1)=fm1(j)*u(i,j)
            m4(i,j)=t1
            m4(i,j+1)=t2
         END DO
      END DO   
        
      CALL VRFFTB(m+1,n,m3(0,1),fftwork,syp1,wsave)
      CALL VRFFTB(m+1,n,m4(0,1),fftwork,syp1,wsave) 
      
      DO j=1,n
         DO i=0,m
            nlu(i,j)=m1(i,j)*m3(i,j)+m2(i,j)*m4(i,j)
         END DO
      END DO                 

C m3=physical(v_x), m4=physical(v_y)

C j=1 and j=n
      DO i=0,m
         t1=0.0D+0
         t2=0.0D+0
         DO k=0,m
            t1=t1+D1(i,k)*v(k,1)
            t2=t2+D1(i,k)*v(k,n) 
         END DO
         m3(i,1)=0.0D+0
         m3(i,n)=0.0D+0
         m4(i,1)=t1
         m4(i,n)=t2
      END DO
      
C j=2,...,n-1
      
      DO j=2,n-2,2
         DO i=0,m
            t1=0.0D+0
            t2=0.0D+0
            DO k=0,m
               t1=t1+D1(i,k)*v(k,j)
               t2=t2+D1(i,k)*v(k,j+1) 
            END DO
            m3(i,j)=fm1(j+1)*v(i,j+1)
            m3(i,j+1)=fm1(j)*v(i,j)
            m4(i,j)=t1
            m4(i,j+1)=t2
         END DO
      END DO   
        
      CALL VRFFTB(m+1,n,m3(0,1),fftwork,syp1,wsave)
      CALL VRFFTB(m+1,n,m4(0,1),fftwork,syp1,wsave) 
      
      DO j=1,n
         DO i=0,m
            nlv(i,j)=m1(i,j)*m3(i,j)+m2(i,j)*m4(i,j)
         END DO
      END DO 

C transform NLTs

      CALL VRFFTF(m+1,n,nlu(0,1),fftwork,syp1,wsave)
      CALL VRFFTF(m+1,n,nlv(0,1),fftwork,syp1,wsave)            

      RETURN
      END    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      REAL*8 FUNCTION TF1(time)
      IMPLICIT NONE

      REAL*8 time

      tf1=DCOS(time)

      RETURN
      END
!!!!!!!!!!!!!!!!!!!!!
      REAL*8 FUNCTION TF2(time)
      IMPLICIT NONE

      REAL*8 time

      tf2=-DSIN(time)

      RETURN
      END
!!!!!!!!!!!!!!!!!!!!!
      REAL*8 FUNCTION TF3(time)
      IMPLICIT NONE

      REAL*8 time

      tf3=DCOS(time)*DCOS(time)

      RETURN
      END
!!!!!!!!!!!!!!!!!!!!!
      SUBROUTINE CHEB(x,D1,D2,m,arrdim)
      IMPLICIT NONE 
      
C input/output variables
     
      INTEGER m,arrdim
      REAL*8  x(arrdim+1),D1(arrdim+1,arrdim+1),D2(arrdim+1,arrdim+1)

C local parameters

      INTEGER size
      PARAMETER(size=65)  
      REAL*8 pi
      PARAMETER(pi=3.14159265358979323846264338327950288D+0)

C local variables           
           
      INTEGER i,j,k,n,n1,n2
      REAL*8  th(size),z(size,size),dx(size,size)
      REAL*8  cvec(size),c(size,size),sum
      
C executable code
      
      IF (size .LT. m+1) THEN
         WRITE(6,*) 'IN CHEB: SIZE is too small!!'
         WRITE(6,*) 'Must be at least ',m+1
         STOP
      END IF    
       
      n=m+1

      j=m
      DO i=1,n
         th(i)=DBLE(i-1)*pi/DBLE(m)
         x(i)=DSIN(DBLE(j)*pi/DBLE(2*m))
         j=j-2
      END DO     

      DO i=1,n
         DO j=1,n
            z(i,j)=2.0D+0*DSIN(th(j)/2.0D+0+th(i)/2.0D+0)*
     &                    DSIN(th(j)/2.0D+0-th(i)/2.0D+0)
         END DO
      END DO

      n1=n/2
      n2=n-n1 
      DO j=1,n
         DO i=1,n1
            dx(i,j)=z(i,j)
         END DO
         DO i=1,n2
            dx(n1+i,j)=-z(n2+1-i,n+1-j)
         END DO
         dx(j,j)=1.0D+0
      END DO         
      DO j=1,n
         DO i=1,n
            z(i,j)=1.0D+0/dx(i,j)
         END DO
         z(j,j)=0.0D+0
      END DO
   
      cvec(1)=1.0D+0
      DO i=2,n
         cvec(i)=-cvec(i-1)
      END DO
      DO i=1,n
         k=1
         DO j=i,n
            c(i,j)=cvec(k)
            k=k+1
         END DO 
         k=2
         DO j=i-1,1,-1
            c(i,j)=cvec(k)
            k=k+1
         END DO
      END DO 
      DO i=2,n-1
         c(i,1)=c(i,1)/2.0D+0
         c(i,n)=c(i,n)/2.0D+0
      END DO       
      DO j=2,n-1
         c(1,j)=2.0D+0*c(1,j)
         c(n,j)=2.0D+0*c(n,j)
      END DO
 
      
C compute D1

      DO j=1,n
         DO i=1,n
            D1(i,j)=z(i,j)*c(i,j)
         END DO
         D1(j,j)=z(j,j)*(c(j,j)-1.0D+0)
      END DO
      DO i=1,n
         sum=0.0D+0
         DO j=1,n
            sum=sum+D1(i,j)
         END DO
         D1(i,i)=-sum
      END DO

C compute D2

      DO j=1,n
         DO i=1,n
            D2(i,j)=2.0D+0*z(i,j)*(c(i,j)*D1(i,i)-D1(i,j))
         END DO
      END DO
      DO i=1,n
         sum=0.0D+0
         DO j=1,n
            sum=sum+D2(i,j)
         END DO
         D2(i,i)=-sum
      END DO

      RETURN
      END
